/*
 *  Copyright (C) Chaperon. All rights reserved.
 *  -------------------------------------------------------------------------
 *  This software is published under the terms of the Apache Software License
 *  version 1.1, a copy of which has been included  with this distribution in
 *  the LICENSE file.
 */

package net.sourceforge.chaperon.cocoon;

import org.apache.avalon.excalibur.pool.Poolable;
import org.apache.avalon.excalibur.pool.Recyclable;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.Composable;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.ParameterException;
import org.apache.avalon.framework.parameters.Parameters;

import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceException;
import org.apache.excalibur.source.SourceValidity;

import org.apache.cocoon.Constants;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.components.source.SourceUtil;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.transformation.AbstractTransformer;
import org.apache.cocoon.util.HashUtil;
import org.apache.cocoon.xml.EmbeddedXMLPipe;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.ContentHandler;
import org.xml.sax.helpers.AttributesImpl;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Enumeration;

import net.sourceforge.chaperon.grammar.Grammar;
import net.sourceforge.chaperon.grammar.SyntaxErrorException;
import net.sourceforge.chaperon.grammar.generator.SAXGrammarGenerator;
import net.sourceforge.chaperon.parser.ParserException;
import net.sourceforge.chaperon.parser.Parser;
import net.sourceforge.chaperon.parser.ParserTable;
import net.sourceforge.chaperon.parser.generator.ParserTableGenerator;
import net.sourceforge.chaperon.parser.output.EventQueue;
import net.sourceforge.chaperon.parser.output.SAXEventAdapter;

/**
 * The FragmentTextParser parse text in a XML document
 *
 * @author Stephan Michels
 * @version CVS $Id: FragmentTextParserTransformer.java,v 1.14 2002/06/01 19:43:16 benedikta Exp $
 */
public class FragmentTextParserTransformer extends AbstractTransformer
        implements Composable, Recyclable, Disposable, /*Poolable,*//* Configurable,*/
                   CacheableProcessingComponent
{
  private static String EXTRACT_URI = "http://chaperon.sourceforge.net/schema/textfragment/1.0";
  private static String EXTRACT_ELEMENT = "textfragment";

  /** The URI of the parser exception DTD */
  public final static String URI = "http://chaperon.sourceforge.net/schema/parser-exception/1.0";

  /** Element name */
  public final static String ELEMENT = "parser-exception";

  /** Attribute name of the message property */
  public final static String MESSAGE_ATTRIBUTE = "message";

  /** Attribute name of the message property */
  public final static String LINENUMBER_ATTRIBUTE = "linenr";

  /** Attribute name of the message property */
  public final static String COLUMNNUMBER_ATTRIBUTE = "columnnr";

  /** Element name for acceptedsymbols */
  public final static String ACCEPTEDSYMBOL_ELEMENT = "parser-exception-accept";

  private int              _extractLevel = 0;

  private StringBuffer     _text = null;

  private Source           _grammarSource = null;
  
  private ComponentManager _manager = null;
  private SourceResolver   _resolver = null;
  
  private String           _grammar = null;
  private boolean          _includeIgnorableTokens = false;
  
  private ParserTable      _parsertable = null;

  /**
   * Pass the Configuration to the Configurable  class. This method must 
   * always be called after the constructor and before any other method.
   *
   * @param conf the class configurations.
   *
   * @throws ConfigurationException Thrown when a Configurable component 
   *                                cannot be configured properly, or if 
   *                                a value cannot be retrieved properly.
   */
  /*public void configure(Configuration conf) throws ConfigurationException
  {
    if (conf != null)
    {
      Configuration[] parameter = conf.getChildren("parameter");

      for (int i = 0; i < parameter.length; i++)
        if (parameter[i].getAttribute("name").equals("includeignorabletokens"))
          _includeIgnorableTokens = parameter[i].getAttributeAsBoolean("value", false);
        else if (parameter[i].getAttribute("name").equals("grammar"))
          _grammar = parameter[i].getAttribute("value");
    }
  }*/

  /**
   * Pass the ComponentManager to the composer. The Composable 
   * implementation should use the specified ComponentManager
   * to acquire the components it needs for execution.
   *
   * @param manager The ComponentManager which this Composable uses.
   */
  public void compose(ComponentManager manager)
  {
    _manager = manager;
  }

  /**
   * Recycle this component.
   * All instance variables are set to <code>null</code>.
   */
  public void recycle()
  {
    if ((_resolver!=null) && (_grammarSource!=null))
    {
      _resolver.release(_grammarSource);
      _grammarSource = null;
    }

    _parsertable = null;
  }

  /**
   * The dispose operation is called at the end of a 
   * components lifecycle.
   */
  public void dispose()
  {
    if ((_resolver!=null) && (_grammarSource!=null))
    {
      _resolver.release(_grammarSource);
      _grammarSource = null;
    }

    _manager = null;
  }

  /**
   * Set the SourceResolver, objectModel Map, the source and sitemap 
   * Parameters used to process the request.
   *
   * @param resolver Source resolver
   * @param objectmodel Object model
   * @param src Source
   * @param parameters Parameters
   */
  public void setup(SourceResolver resolver, Map objectmodel, String src,
                    Parameters parameters)
    throws ProcessingException, SAXException, IOException
  {
    //super.setup(resolver, objectmodel, src, parameters);
    _resolver = resolver;
    ParserTableStore store = null;
    try
    {
      //getLogger().debug("Try to resolve "+src);

      _includeIgnorableTokens = parameters.getParameterAsBoolean("includeignorabletokens", false);
      _grammar = parameters.getParameter("grammar");

      /*System.out.println("includeIgnorableTokens="+_includeIgnorableTokens);
      System.out.println("grammar="+_grammar);*/

      store = (ParserTableStore) _manager.lookup(ParserTableStore.ROLE);

      _grammarSource = resolver.resolveURI(_grammar);

      /*System.out.println("_grammarSource.getSystemId()="+_grammarSource.getSystemId());
      System.out.println("_grammarSource.getValidity()="+_grammarSource.getValidity());
      System.out.println("_grammarSource.getClass()="+_grammarSource.getClass());
      if (store.hasParserTable(_grammarSource.getSystemId()))
        System.out.println("store.getParserTableValidity(_grammarSource.getSystemId())="+
                           store.getParserTableValidity(_grammarSource.getSystemId()));*/

      /*System.out.println("Class of source "+_grammarSource.getClass().getName());
      System.out.println("Validity for "+_grammarSource.getSystemId()+" from the entry in the store:"+
                        store.getParserTableValidity(_grammarSource.getSystemId()));
      System.out.println("Validity from the source "+_grammarSource.getSystemId()+":"+
                        _grammarSource.getValidity());*/

      if ((!store.hasParserTable(_grammarSource.getSystemId())) ||
          (store.getParserTableValidity(_grammarSource.getSystemId())==null) ||
          (!store.getParserTableValidity(_grammarSource.getSystemId()).isValid(_grammarSource.getValidity())))
      {
        getLogger().debug("(Re)building the parsertable from '"+_grammarSource.getSystemId()+"'");

        SAXGrammarGenerator grammargenerator = new SAXGrammarGenerator();
        SourceUtil.toSAX(_grammarSource, grammargenerator, _manager);

        Grammar grammar = grammargenerator.getGrammar();
        SyntaxErrorException see = grammar.validate();
        if (see!=null)
        { 
          getLogger().error("Grammar is not correct", see);
          throw new ProcessingException("Grammar is not correct", see);
        }

        ParserTableGenerator parsertablegenerator = new ParserTableGenerator(grammar);
        _parsertable = parsertablegenerator.getParserTable();

        store.store(_grammarSource.getSystemId(), _parsertable, _grammarSource.getValidity());
      }
      else
      {
        getLogger().debug("Getting parsertable from store for '"+_grammarSource.getSystemId()+"'");
        _parsertable = store.getParserTable(_grammarSource.getSystemId());
      }

    } catch (ParameterException pe)
    {
      getLogger().error("Error during retrieving a parameter", pe);
      throw new ProcessingException("Error during retrieving a parameter", pe);
    } catch (SourceException se)
    {
      getLogger().error("Error during resolving of '" + src + "'.", se);
      throw new ProcessingException("Error during resolving of '" + src + "'.", se);
    } catch (ComponentException ce)
    {
      getLogger().error("Could not lookup for component", ce);
      throw new ProcessingException("Could not lookup for component", ce);
    } catch (SyntaxErrorException see)
    {
      getLogger().error("Grammar is not correct", see);
      throw new ProcessingException("Grammar is not correct", see);
    } finally
    {
      if (store!=null)
        _manager.release(store);
    }
  }

  /**
   * Generate the unique key.
   * This key must be unique inside the space of this component.
   *
   * @return The generated key hashes the src
   */
  public Serializable generateKey() 
  {
    return _grammarSource.getSystemId();
  }

  /**
   * Generate the validity object.
   *
   * @return The generated validity object or <code>null</code> if the
   *         component is currently not cacheable.
   */
  public SourceValidity generateValidity() 
  {
    return _grammarSource.getValidity();
  }

  /**
   * Receive notification of the beginning of an element.
   *
   * @param uri The Namespace URI, or the empty string if the element has no
   *            Namespace URI or if Namespace
   *            processing is not being performed.
   * @param loc The local name (without prefix), or the empty string if
   *            Namespace processing is not being performed.
   * @param raw The raw XML 1.0 name (with prefix), or the empty string if
   *            raw names are not available.
   * @param a The attributes attached to the element. If there are no
   *          attributes, it shall be an empty Attributes object.
   */
  public void startElement(String uri, String loc, String raw,
                           Attributes a) throws SAXException
  {
    if (this.EXTRACT_URI.equals(uri) && this.EXTRACT_ELEMENT.equals(loc)
            && (_grammar != null))
    {
      _extractLevel++;
      _text = new StringBuffer();

      /*if (extractLevel==1)
      {
        String grammar = a.getValue("grammar");
        if ((grammar!=null) && (grammar.length()>0))
          this.grammar = grammar;
      }*/
    }
    else
      super.startElement(uri, loc, raw, a);
  }

  /**
   * Receive notification of the end of an element.
   *
   * @param uri The Namespace URI, or the empty string if the element has no
   *            Namespace URI or if Namespace
   *            processing is not being performed.
   * @param loc The local name (without prefix), or the empty string if
   *            Namespace processing is not being performed.
   * @param raw The raw XML 1.0 name (with prefix), or the empty string if
   *            raw names are not available.
   */
  public void endElement(String uri, String loc,
                         String raw) throws SAXException
  {
    if (this.EXTRACT_URI.equals(uri) && this.EXTRACT_ELEMENT.equals(loc)
            && (_grammar != null))
    {
      _extractLevel--;
  
      //InputSource inputsource = new InputSource(new ByteArrayInputStream(text.toString().getBytes()));
  
      TextParser parser = null;
      try
      {
        parser = (TextParser) _manager.lookup("net.sourceforge.chaperon.cocoon.TextParser");
        EventQueue queue = parser.parse(_parsertable, new ByteArrayInputStream(_text.toString().getBytes()));

        SAXEventAdapter adapter = new SAXEventAdapter(super.contentHandler, _includeIgnorableTokens, true);
        //adapter.enableLogging(getLogger());
        queue.fireEvents(adapter);
      }
      catch (ComponentException ce)
      {
        getLogger().error("Could not lookup for component", ce);
        throw new SAXException("Could not lookup for component", ce);
      }
      /*catch (SyntaxErrorException ser)
      {
        getLogger().error("Grammar is not valid", ser);
        throw new SAXException("Grammar is not valid", ser);
      }*/
      catch (ParserException pe)
      {
        getLogger().error("Document ist not valid", pe);
        EmbeddedXMLPipe pipe = new EmbeddedXMLPipe(super.contentHandler);
        toSAX(pipe, pe);
      }
      finally
      {
        if (parser!=null)
          _manager.release(parser);
      }  

      _text = null;
    }
    else
      super.endElement(uri, loc, raw);
  }

  /**
   * Receive notification of character data.
   *
   * @param c The characters from the XML document.
   * @param start The start position in the array.
   * @param len The number of characters to read from the array.
   */
  public void characters(char c[], int start, int len) throws SAXException
  {
    if (_extractLevel > 0)
      _text.append(c, start, len);
    else
      super.characters(c, start, len);
  }

  /** 
   * Serialize the exception to a SAX stream
   *
   * @param handler The content handler, which receives the SAX events
   *
   * @throws SAXException
   */
  public void toSAX(ContentHandler handler, ParserException pe) throws SAXException
  {
    handler.startDocument();

    AttributesImpl attributes = new AttributesImpl();

    attributes.addAttribute(URI, MESSAGE_ATTRIBUTE, MESSAGE_ATTRIBUTE,
                            "CDATA", pe.getMessage());
    attributes.addAttribute(URI, LINENUMBER_ATTRIBUTE, LINENUMBER_ATTRIBUTE,
                            "CDATA", String.valueOf(pe.getLineNumber()));
    attributes.addAttribute(URI, COLUMNNUMBER_ATTRIBUTE, COLUMNNUMBER_ATTRIBUTE,
                            "CDATA", String.valueOf(pe.getColumnNumber()));

    handler.startElement(URI, ELEMENT, ELEMENT, attributes);

    for(int i=0; i<pe.getAcceptedSymbols().length; i++)
    {
      handler.startElement(URI, ACCEPTEDSYMBOL_ELEMENT, ACCEPTEDSYMBOL_ELEMENT, new AttributesImpl());
      handler.characters(pe.getAcceptedSymbols()[i].toCharArray(), 0, pe.getAcceptedSymbols()[i].length());
      handler.endElement(URI, ACCEPTEDSYMBOL_ELEMENT, ACCEPTEDSYMBOL_ELEMENT);
    }

    handler.endElement(URI, ELEMENT, ELEMENT);

    handler.endDocument();
  }
}
